;; Generated package description from transient-cycles.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "transient-cycles" "2.0" "Define command variants with transient cycling" '((emacs "29.1")) :commit "11e547c11b84fa81bb4ef8ec33ee777f96576e8c" :authors '(("Sean Whitton" . "spwhitton@spwhitton.name")) :maintainer '("Sean Whitton" . "spwhitton@spwhitton.name") :keywords '("buffer" "window" "processes" "minor-mode" "convenience") :url "https://git.spwhitton.name/dotfiles/tree/.emacs.d/site-lisp/transient-cycles.el")
